/*
 * array.c - implementation of the intrinsic array type. See also array.h.
 */
#include "ptr.h"
#include "exec.h"
#include "op.h"
#include "int.h"
#include "buf.h"
#include "primes.h"

/*
 * Function to do the hard work for the macro pushcheck(). See array.h.
 */
int
ici_growarray(array_t *a, ptrdiff_t n)
{
    register object_t	**e;

    if (objof(a)->o_flags & O_ATOM)
    {
	error = "attempt to modify an atomic array";
	return 1;
    }
    if (a->a_base == NULL)
    {
	/*
	 * Virgin array, first memory allocation (only happens because
	 * of special case in ici_evaluate, arrays made with new_array()
	 * would have already have had this work done).
	 */
	if (n < FEW_OBJS)
	    n = FEW_OBJS;
	if ((e = (object_t **)ici_alloc(n * sizeof(object_t *))) == NULL)
	    return 1;
	a->a_top = e;
	a->a_base = e;
	a->a_limit = e + n;
    }
    else
    {
	/*
	 * Append space to the array.  We don't use realloc to ensure
	 * that memory exhaustion is cleanly recovereable.
	 */
	if ((a->a_limit - a->a_base) * 3 / 2 < a->a_limit - a->a_base + n)
	    n += (a->a_limit - a->a_base) + 10;
	else
	    n = (a->a_limit - a->a_base) * 3 / 2;
	if ((e = (object_t **)ici_alloc(n * sizeof(object_t *))) == NULL)
	    return 1;
	memcpy((char *)e, (char *)a->a_base,
	    (a->a_limit - a->a_base) * sizeof(object_t *));
	a->a_top = e + (a->a_top - a->a_base);
	ici_free((char *)a->a_base);
	a->a_base = e;
	a->a_limit = e + n;
    }
    return 0;
}

/*
 * Function to do the hard work for the macro arrayprobe(). See array.h.
 */
int
ici_faultarray(array_t *a, ptrdiff_t i)
{
    if (objof(a)->o_flags & O_ATOM)
    {
	error = "attempt to modify an atomic array";
	return 1;
    }
    if (i < 0)
    {
	sprintf(buf, "attempt to write at array index %ld\n", (long)i);
	error = buf;
	return 1;
    }
    ++i;
    i -= a->a_top - a->a_base;
    if (pushcheck(a, i))
	return 1;
    while (--i >= 0)
	*a->a_top++ = objof(&o_null);
    return 0;
}

/*
 * Function used in fail case of the macro popcheck(). See array.h.
 */
int
ici_badindex(void)
{
    error = "array index range error";
    return 1;
}

/*
 * Return a new array.  It will have room for at least FEW_OBJS elements at
 * the start. Has ref count 1. Returns NULL on failure, usual conventions.
 */
array_t *
new_array(void)
{
    register array_t	*a;

    if ((a = talloc(array_t)) == NULL)
	return NULL;
    objof(a)->o_type = &ici_array_type;
    objof(a)->o_tcode = TC_ARRAY;
    objof(a)->o_flags = 0;
    objof(a)->o_nrefs = 1;
    rego(a);
    a->a_base = NULL;
    a->a_top = NULL;
    a->a_limit = NULL;
    /*
     * Note that the following FEW_OBJS is choosen to correspond with the
     * size of the smaller special memory list.
     */
    if ((a->a_base = (object_t **)ici_alloc(FEW_OBJS*sizeof(object_t *))) == NULL)
    {
	decref(a);
	return NULL;
    }
    a->a_top = a->a_base;
    a->a_limit = a->a_base + FEW_OBJS;
    return a;
}

/*
 * Mark this and referenced unmarked objects, return memory costs.
 * See comments on t_mark() in object.h.
 */
static unsigned long
mark_array(object_t *o)
{
    object_t		**e;
    unsigned long	mem;

    o->o_flags |= O_MARK;
    mem = sizeof(array_t)
    	+ (arrayof(o)->a_limit - arrayof(o)->a_base) * sizeof(object_t *);
    for (e = arrayof(o)->a_base; e < arrayof(o)->a_top; ++e)
	mem += mark(*e);
    return mem;
}

/*
 * Free this object and associated memory (but not other objects).
 * See the comments on t_free() in object.h.
 */
static void
free_array(object_t *o)
{
    if (arrayof(o)->a_base != NULL)
	ici_free(arrayof(o)->a_base);
    /*
     * This special guard is only needed for arrays because the execution
     * loop uses static arrays which get their memory freed by this
     * routine, but don't have allocated bodies.
     */
    if (o->o_nrefs == 0)
	ici_free(o);
}

/*
 * Returns 0 if these objects are equal, else non-zero.
 * See the comments on t_cmp() in object.h.
 */
static int
cmp_array(object_t *o1, object_t *o2)
{
    int		i;

    if (o1 == o2)
	return 0;
    i = arrayof(o1)->a_top - arrayof(o1)->a_base;
    if (arrayof(o2)->a_top - arrayof(o2)->a_base != i)
	return 1;
    return memcmp
    (
	arrayof(o1)->a_base,
	arrayof(o2)->a_base,
	i * sizeof(object_t *)
    );
}

/*
 * Return a copy of the given object, or NULL on error.
 * See the comment on t_copy() in object.h.
 */
static object_t *
copy_array(object_t *o)
{
    array_t	*na;
    ptrdiff_t	n;

    if ((na = new_array()) == NULL)
	return NULL;
    n = arrayof(o)->a_top - arrayof(o)->a_base;
    if (pushcheck(na, n))
	goto fail;
    memcpy(na->a_base, arrayof(o)->a_base, n * sizeof(object_t *));
    na->a_top += n;
    return objof(na);

fail:
    decref(na);
    return NULL;
}

/*
 * Return a hash sensitive to the value of the object.
 * See the comment on t_hash() in object.h
 */
static unsigned long
hash_array(object_t *o)
{
    unsigned long		h;
    unsigned long const		*p;
    object_t			**po;
    static unsigned long const	primes[] = {0x892D, 0xB91D, 0xB345, 0x254F, 0x67E5};

    h = ARRAY_PRIME;
    p = primes;
    po = arrayof(o)->a_base;
    while (po < arrayof(o)->a_top)
    {
	h += ((unsigned long)*po++ >> 4) * *p++;
	if (p == &primes[nels(primes)])
	    p = primes;
    }
    return h;
}

/*
 * Assign to key k of the object o the value v. Return 1 on error, else 0.
 * See the comment on t_assign() in object.h.
 *
 * The key k must be a positive integer. The array will attempt to grow
 * to accomodate the new index as necessary.
 */
static int
assign_array(object_t *o, object_t *k, object_t *v)
{
    long	i;

    if (!isint(k))
	return assign_simple(o, k, v);
    i = intof(k)->i_value;
    if (arrayprobe(arrayof(o), (ptrdiff_t)i))
	return 1;
    arrayof(o)->a_base[i] = v;
    return 0;
}

/*
 * Return the object at key k of the obejct o, or NULL on error.
 * See the comment on t_fetch in object.h.
 *
 * The key k must be an integer. Reading non existing keys results in
 * an NULL value (not an error).
 */
static object_t *
fetch_array(object_t *o, object_t *k)
{
    long	i;

    if (!isint(k))
	return fetch_simple(o, k);
    i = intof(k)->i_value;
    if (i >= 0 && i < arrayof(o)->a_top - arrayof(o)->a_base)
	return arrayof(o)->a_base[i];
    return objof(&o_null);
}

/*
 * mark any any... => array (os)
 */
int
ici_op_offsq()
{
    array_t	*a;
    ptrdiff_t	i;

    for (i = 1; os->a_top - i > os->a_base; ++i)
    {
	if (o_top[-i] == objof(&o_mark))
	    break;
    }
    if (o_top[-i] != objof(&o_mark))
	return ici_badindex();
    --i;
    if ((a = new_array()) == NULL || pushcheck(a, i))
	return 1;
    memcpy(a->a_base, &o_top[-i], i * sizeof(object_t *));
    a->a_top += i;
    o_top -= i;
    o_top[-1] = objof(a);
    decref(a);
    --x_top;
    return 0;
}

/*
 * obj => array 0 (the array contains the obj)
 */
int
ici_op_mklvalue()
{
    array_t	*a;

    if ((a = new_array()) == NULL)
	return 1;
    *a->a_top++ = o_top[-1];
    o_top[-1] = objof(a);
    *o_top++ = objof(o_zero);
    decref(a);
    --x_top;
    return 0;
}

type_t	ici_array_type =
{
    mark_array,
    free_array,
    hash_array,
    cmp_array,
    copy_array,
    assign_array,
    fetch_array,
    "array"
};

op_t	o_offsq		= {OBJ(TC_OP, op_type), ici_op_offsq};
op_t	o_mklvalue	= {OBJ(TC_OP, op_type), ici_op_mklvalue};
