#include "buf.h"
#include "exec.h"
#include "func.h"
#include "int.h"
#include "float.h"
#include "str.h"
#include "null.h"
#include "op.h"
#include "catch.h"

/*
 * The common code used by ici_func() and ici_call() below. Also a
 * varient of ici_func() taking a variable argument list. See comment
 * on ici_func() below for details.
 */
char *
ici_funcv(object_t *func_obj, char *types, va_list va)
{
    object_t		*call_obj;
    catch_t		*frame;
    int			nargs;
    int			arg;
    object_t            *member_obj;
    object_t		*ret_obj;
    char		ret_type;
    char		*ret_ptr;
    int			os_depth;

    if (types[0] != '\0' && types[1] == '@')
    {
        member_obj = va_arg(va, object_t *);
        types += 2;
    }
    else
    {
        member_obj = NULL;
    }

    if (types[0] != '\0' && types[1] == '=')
    {
	ret_type = types[0];
	ret_ptr = va_arg(va, char *);
	types += 2;
    }
    else
    {
	ret_type = '\0';
	ret_ptr = NULL;
    }

    os_depth = o_top - os->a_base;
    call_obj = NULL;
    /*
     * We include an extra 80 in our pushcheck, see start of evaluate().
     */
    nargs = strlen(types);
    if (pushcheck(os, nargs + 80))
	return error;
    for (arg = 0; arg < nargs; ++arg)
	*o_top++ = objof(&o_null);
    for (arg = -1; arg >= -nargs; --arg)
    {
	switch (*types++)
	{
	case 'o':
	    o_top[arg] = va_arg(va, object_t *);
	    break;

	case 'i':
	    if ((o_top[arg] = objof(new_int(va_arg(va, long)))) == NULL)
		goto fail;
	    decref(o_top[arg]);
	    break;

	case 'q':
	    o_top[arg] = objof(&o_quote);
	    --nargs;
	    break;

	case 's':
	    if ((o_top[arg] = objof(new_cname(va_arg(va, char *)))) == NULL)
		goto fail;
	    decref(o_top[arg]);
	    break;

	case 'f':
	    if ((o_top[arg] = objof(new_float(va_arg(va, double)))) == NULL)
		goto fail;
	    decref(o_top[arg]);
	    break;

	default:
	    error = "error in function call";
	    goto fail;
	}
    }
    if (member_obj != NULL)
    {
        *o_top++ = member_obj;
        nargs++;
    }
    *o_top++ = func_obj;
    if ((call_obj = objof(new_op(NULL, OP_CALL, nargs))) == NULL)
	goto fail;
    if ((frame = new_catch(NULL, os_depth, v_top - vs->a_base)) == NULL)
	goto fail;
    if ((ret_obj = ici_evaluate(objof(call_obj), frame)) == NULL)
	goto fail;
    switch (ret_type)
    {
    case '\0':
	decref(ret_obj);
	break;

    case 'o':
	*(object_t **)ret_ptr = ret_obj;
	break;

    case 'i':
	if (!isint(ret_obj))
	    goto typeclash;
	*(long *)ret_ptr = intof(ret_obj)->i_value;
	decref(ret_obj);
	break;

    case 'f':
	if (!isfloat(ret_obj))
	    goto typeclash;
	*(double *)ret_ptr = floatof(ret_obj)->f_value;
	decref(ret_obj);
	break;

    case 's':
	if (!isstring(ret_obj))
	    goto typeclash;
	*(char **)ret_ptr = stringof(ret_obj)->s_chars;
	decref(ret_obj);
	break;

    default:
    typeclash:
	decref(ret_obj);
	error = "incorrect return type";
	goto fail;
    }
    decref(call_obj);
    return NULL;

fail:
    if (call_obj != NULL)
	decref(call_obj);
    o_top = os->a_base + os_depth;
    return error;
}

/*
 * Varient of ici_call() (see below) taking a variable argument list.
 */
char *
ici_callv(char *func_name, char *types, va_list va)
{
    object_t		*name_obj;
    object_t		*func_obj;
    object_t            *member_obj;
    char		*result;

    name_obj = NULL;
    func_obj = NULL;

    if ((name_obj = objof(new_cname(func_name))) == NULL)
	return error;
    if (types[0] != '\0' && types[1] == '@')
    {
        va_list tmp;
        tmp = va;
        member_obj = va_arg(tmp, object_t *);
        if ((func_obj = fetch(member_obj, name_obj)) == objof(&o_null))
        {
	    sprintf(buf, "\"%s\" undefined in object", func_name);
	    error = buf;
	    decref(name_obj);
	    return error;
        }
        va_end(tmp);
    }
    else if ((func_obj = fetch(v_top[-1], name_obj)) == objof(&o_null))
    {
	sprintf(buf, "\"%s\" undefined", func_name);
	error = buf;
	decref(name_obj);
	return error;
    }
    decref(name_obj);
    name_obj = NULL;
    result = ici_funcv(func_obj, types, va);
    return result;
}

/*
 * ici_func(func, types, args...)
 *
 * Call an ICI function from C with simple argument types and return value.
 *
 * Types can be of the forms ".=..." or "...".  In the first case the 1st
 * extra arg is used as a pointer to store the return value through.
 *
 * Type key letters are:
 * 	i	a long
 *	f 	a double
 *	s	a '\0' terminated string
 *	o	an ici object
 *
 * When a string is returned it is a pointer to the character data of an
 * internal ICI string object. It will only remain valid until the next
 * call to any ICI function.  When an object is returned it has been incref'ed
 * (i.e. it is held against garbage collection).
 */
char *
ici_func(object_t *func_obj, char *types, ...)
{
    va_list	va;
    char	*result;

    va_start(va, types);
    result = ici_funcv(func_obj, types, va);
    va_end(va);
    return result;
}

/*
 * ici_call(name, types, args...)
 *
 * Call an ici function by name from C with simple argument types and
 * return value.  The named is looked up in the current scope or in 
 * a supplied struct.
 *
 * Types can be of the forms ".=...", ".@.=...", ".@..." or "...".
 * In the first case the 1st extra arg is used as a pointer to store the
 * return value through. In the second and third cases the first extra 
 * arg is a struct from which the function is fetched; this object 
 * will be passed as the first parameter to the function. In the second case
 * the second extra arg will be used for the return value.
 *
 * Type key letters are:
 * 	i	a long
 *	f 	a double
 *	s	a '\0' terminated string
 *	o	an ici object
 *
 * When a string is returned it is a pointer to the character data of an
 * internal ICI string object. It will only remain valid until the next
 * call to any ICI function.  When an object is returned it has been incref'ed
 * (i.e. it is held against garbage collection).
 */
char *
ici_call(char *func_name, char *types, ...)
{
    object_t		*name_obj;
    object_t		*func_obj;
    object_t            *member_obj;
    va_list		va;
    char		*result;

    name_obj = NULL;
    func_obj = NULL;

    if ((name_obj = objof(new_cname(func_name))) == NULL)
	return error;
    if (types[0] != '\0' && types[1] == '@')
    {
        va_start(va, types);
        member_obj = va_arg(va, object_t *);
        if ((func_obj = fetch(member_obj, name_obj)) == objof(&o_null))
        {
	    sprintf(buf, "\"%s\" undefined in object", func_name);
	    error = buf;
	    decref(name_obj);
	    return error;
        }
    }
    else if ((func_obj = fetch(v_top[-1], name_obj)) == objof(&o_null))
    {
	sprintf(buf, "\"%s\" undefined", func_name);
	error = buf;
	decref(name_obj);
	return error;
    }
    va_start(va, types);
    decref(name_obj);
    name_obj = NULL;
    result = ici_funcv(func_obj, types, va);
    va_end(va);
    return result;
}
