#include "exec.h"
#include "op.h"
#include "catch.h"
#include "ptr.h"
#include "func.h"
#include "str.h"
#include "buf.h"
#include "pc.h"
#include "int.h"
#include "struct.h"
#include "set.h"
#include "parse.h"
#include "float.h"
#include "re.h"
#include "src.h"
#include "null.h"
#include "forall.h"
#include "primes.h"
#ifndef NOTRACE
#include "trace.h"
#endif

exec_t          ex;             /* See #defines in exec.h */

int_t           *o_zero;
int_t           *o_one;

/*
 * Set this to non-zero to cause an "aborted" failure even when the ICI
 * program is stuck in an infinite loop.  For use in embeded systems etc
 * to provide protection against badly behaved user programs.  How it gets
 * set is not addressed here (by an interupt perhaps).  Remember to clear
 * it before re-running any ICI code.
 */
volatile int            ici_aborted;

#ifndef NODEBUGGING
/*
 * When non-zero this enables calls to the debugger functions (see idb2.c).
 */
extern int      ici_debug_enabled;
#endif

/*
 * Initialise the global exec structure. The given scope is the structure
 * for autos, and its super for statics.
 */
int
init_exec(struct_t *scope)
{
    /*
     * The exec structure contains in-line array_t's (i.e. not pointers).
     * This is so that the tops of the operation and execution stacks are
     * at fixed absolute memory addresses (to save indirections).  We
     * initialise these arrays here and incref() them.  They remain incref()
     * to avoid anybody trying to free them.
     */
    objof(xs)->o_type = &ici_array_type;
    objof(xs)->o_tcode = TC_ARRAY;
    objof(xs)->o_nrefs = 1;
    rego(xs);
    if (pushcheck(xs, 80))
        return 1;
    objof(os)->o_type = &ici_array_type;
    objof(os)->o_tcode = TC_ARRAY;
    objof(os)->o_nrefs = 1;
    rego(os);
    if (pushcheck(os, 80))
        return 1;
    objof(vs)->o_type = &ici_array_type;
    objof(vs)->o_tcode = TC_ARRAY;
    objof(vs)->o_nrefs = 1;
    rego(vs);
    if (pushcheck(vs, 80))
        return 1;
    *v_top++ = objof(scope);
    ex.x_count = 20;

    return 0;
}

/*
 * Execute 'code' (any object, normally an array_t of code or a parse_t).
 * The execution procedes on top of the current stacks (execution, operand
 * and variable). This call to evaluate will return when the execution
 * stack again returns to the level it was when entered. It then returns
 * the object left on the operand stack, or &o_null if there wasn't one.
 * The returned object is incref().  Returns NULL on error, usual conventions
 * (i.e. 'error' points to the error message).
 *
 * The execution loop knows when the execution stack returns to its
 * origional level because it puts a catch_t object on it. This object
 * also records the levels of the other two stacks that match. The catch
 * object can be passed in if stacks have been pre-configured specially.
 * But it is normally NULL.
 *
 * This is the main execution loop.  All of the nasty optimisations are
 * concentrated here.  It used to be clean, elegant and 20 lines long.
 * Now it goes faster.
 *
 * Originally each type had an execution method, which was a function
 * pointer like all the other type specific methods.  This was unrolled
 * into a switch within the main loop for speed and variable optimisation.
 * Then many of the op_type operations got unrolled in the loop in the same
 * way.  Gotos got added to short-circuit some of the steps where possible.
 * Then they got removed, apart from the fail exits.  Then one got added again.
 *
 * Note that binop.h is included half way down this function.
 */
object_t *
ici_evaluate(object_t *code, catch_t *frame)
{
    register object_t   *o;
    src_t               *src;
    static src_t        default_src = {OBJ(TC_OTHER, src_type), 0, NULL};

    if
    (
        frame == NULL
        &&
        (frame = new_catch(NULL, o_top-os->a_base, v_top-vs->a_base)) == NULL
    )
        return NULL;

    *x_top++ = objof(frame);

    if (isarray(code))
    {
        if ((code = objof(new_pc(arrayof(code)))) == NULL)
            return NULL;
    }

    *x_top++ = code;

    /*
     * The execution loop.
     */
    src = &default_src;
    incref(src);
    for (;;)
    {
        if (--ex.x_count == 0)
        {
            /*
             * Ensure that there is enough room on all stacks for at
             * least 20 more worst case operations.  See also f_call().
             */
            if (pushcheck(xs, 80))
                goto fail;
            if (pushcheck(os, 80))
                goto fail;
            if (pushcheck(vs, 80))
                goto fail;
            if (ici_aborted)
            {
                error = "aborted";
                goto fail;
            }
            ex.x_count = 20;
        }

        /*
         * Places which would be inclined to continue in this loop that
         * know that they have not increased any stack depths can just
         * goto this label to avoid the check above.
         */
    stable_stacks_continue:
        /*
         * In principle our execution model is pretty simple. We execute
         * the thing on the top of the execution stack. When that is a pc
         * we push the thing the pc points to (with post increment) onto
         * the execution stack and continue.
         *
         * So technically the following test for a pc on top of the
         * execution stack should just be part of the main switch (as it
         * once was). But 90% of the time we execute: pc then other;
         * pc then other etc... Doing this test here means that normally
         * we run these two operations into one another without going
         * round the main loop. And then, enough of the things we do are
         * operators to make a short-circuit of the switch worth while.
         *
         * Code arrays never contain pcs, so the thing picked up from
         * indirecting the pc is not a pc, so there is no general case
         * in the switch.
         */
        if (ispc(o = x_top[-1]))
        {
            if (pcof(o)->pc_next < pcof(o)->pc_limit)
            {
                object_t   *tmp;

                tmp = *pcof(o)->pc_next++;
                o = tmp;
                if (isop(o))
                    goto an_op;
                *x_top++ = o;
            }
            else
            {
                extern int      ici_op_looper();

                --x_top;
                if (isop(x_top[-1]) && opof(x_top[-1])->op_func == ici_op_looper)
                {
                    /*
                     * End of that code array, but if this is a looper pc,
                     * just rewind it (a sort of in-line expansion of
                     * ici_op_looper).
                     */
                    ++x_top;
                    pcof(o)->pc_next = pcof(o)->pc_code->a_base;
                }
                goto stable_stacks_continue;
            }
        }

        switch (o->o_tcode)
        {
        case TC_SRC:
            decref(src);
            src = srcof(o);
            incref(src);
            --x_top;
#ifndef NODEBUGGING
            if (ici_debug_enabled)
                ici_debug->idbg_src(srcof(o));
#endif
#ifndef NOTRACE
            if (trace_yes && (trace_flags & TRACE_SRC))
            {
                if (src->s_filename == NULL)
                    fprintf(stderr, "%d\n", s->s_lineno);
                else
                    fprintf(stderr, "%s:%d\n", filename->s_chars, s->s_lineno);
            }
#endif
            goto stable_stacks_continue;

        case TC_PARSE:
            if (parse_exec())
                goto fail;
            continue;

        case TC_STRING:
            /*
             * Executing a string is the operation of variable lookup.
             * Look up the value of the string on the execution stack
             * in the current scope and push the value onto the operand
             * stack.
             */
            if (stringof(o)->s_vsver == vsver)
            {
                /*
                 * We know directly where the value is because we have
                 * looked up this name since the last change to the scope
                 * or structures etc.
                 */
                *o_top++ = stringof(o)->s_slot->sl_value;
            }
            else
            {
                struct_t        *where;
                slot_t          *sl;

                /*
                 * General case of name lookup.
                 */
                where = structof(v_top[-1]);
                sl = find_slot(&where, o);
                if (sl->sl_key == NULL)
                {
                    /*
                     * We failed to find that name on first lookup.
                     * Try to load a library of that name and repeat
                     * the lookup before deciding it is undefined.
                     */
                    if (ici_load(stringof(o)) < 0)
                        goto fail;
                    sl = find_slot(&where, o);
                    if (sl->sl_key == NULL)
                    {
                        if (chkbuf(stringof(o)->s_nchars + 20))
                            goto fail;
                        sprintf(buf, "\"%s\" undefined", stringof(o)->s_chars);
                        error = buf;
                        goto fail;
                    }
                }
                stringof(o)->s_vsver = vsver;
                stringof(o)->s_struct = where;
                stringof(o)->s_slot = sl;
                *o_top++ = sl->sl_value;
            }
            --x_top;
            continue;

        case TC_CATCH:
            /*
             * This can either be an error catcher which is being poped
             * off (having done its job, but it never got used) or it
             * can be the guard catch (frame marker) indicating it is time
             * to return.
             *
             * First note the top of the operand stack, if there is anything
             * on it, it becomes the return value, else we return &o_null.
             * The caller knows if there is really a value to return.
             */
            if (catchof(o)->c_odepth < o_top - os->a_base)
                o = o_top[-1];
            else
                o = objof(&o_null);
            if (ici_unwind() == frame)
            {
                /*
                 * This is the base of a call to this function.  It is now
                 * time to return.
                 */
                incref(o);
                decref(src);
                return o;
            }
            goto stable_stacks_continue;

        case TC_FORALL:
            if (exec_forall())
                goto fail;
            continue;

        default:
            *o_top++ = *--x_top;
            continue;

        case TC_OP:
            /*
             * We pre-decrement the execution stack for operators as the
             * operator which is currently on top of the execution stack
             * is known to be pointed to by o, it can't be garbage collected
             * because it is in the code array which is pointed to by the
             * pc, and most of the time we arrive at an_op: which has
             * avoided ever pushing it on the stack anyway. For a few
             * cases below we must restore the stack for consistency in
             * called functions.
             */
            --x_top;
        an_op:
            switch (opof(o)->op_ecode)
            {
            case OP_OTHER:
#ifndef NODEBUGGING
                if (ici_debug_enabled)
                {
                    extern int  op_return();
                    if (opof(o)->op_func == ici_op_return)
                        ici_debug->idbg_fnresult(o_top[-1]);
                }
#endif
                *x_top++ = o;
                if ((*opof(o)->op_func)())
                    goto fail;
                continue;

            case OP_CALL:
                *x_top++ = o;
#ifndef NODEBUGGING
            {
                int intrinsic = objof(o_top[-1])->o_flags & O_CFUNC;
                if (ici_debug_enabled)
                    ici_debug->idbg_fncall(o_top[-1], ARGS(), NARGS());
#endif
                if (ici_op_call())
                    goto fail;
#ifndef NODEBUGGING
                if (ici_debug_enabled && intrinsic)
                    ici_debug->idbg_fnresult(o_top[-1]);
            }
#endif
                continue;

            case OP_QUOTE:
                /*
                 * pc           => pc+1 (xs)
                 *              => *pc (os)
                 */
                *o_top++ = *pcof(x_top[-1])->pc_next++;
                continue;

            case OP_AT:
                /*
                 * obj => obj (os)
                 */
                o_top[-1] = atom(o_top[-1], 0);
                goto stable_stacks_continue;

            case OP_NAMELVALUE:
                /*
                 * pc (xs)      => pc+1 (xs)
                 *              => struct *pc (os)
                 * (Ie. the next thing in the code array is a name, put its
                 * lvalue on the operand stack.)
                 */
                o = *pcof(x_top[-1])->pc_next++;
                if (stringof(o)->s_vsver == vsver)
                {
                    /*
                     * We know directly where the value is because we have
                     * looked up this name since the last change to the scope
                     * or structures etc.
                     */
                    *o_top++ = objof(stringof(o)->s_struct);
                }
                else
                {
                    slot_t              *sl;

                    *o_top++ = v_top[-1];
                    sl = find_slot((struct_t **)&o_top[-1], o);
                    if (sl->sl_key == NULL)
                    {
                        /*
                         * We failed to find that name on first lookup.
                         * Try to load a library of that name and repeat
                         * the lookup before deciding it is undefined.
                         */
                        if (ici_load(stringof(o)) < 0)
                            goto fail;
                        sl = find_slot((struct_t **)&o_top[-1], o);
                        if (sl->sl_key == NULL)
                        {
                            /*
                             * If implicit declaration goes back in as a feature,
                             * use this line of code instead of the next 5.
                             * assign(o_top[-1], o, &o_null);
                             */
                            if (chkbuf(stringof(o)->s_nchars + 20))
                                goto fail;
                            sprintf(buf, "\"%s\" undefined", stringof(o)->s_chars);
                            error = buf;
                            goto fail;
                        }
                    }
                    stringof(o)->s_vsver = vsver;
                    stringof(o)->s_struct = structof(o_top[-1]);
                    stringof(o)->s_slot = sl;
                }
                *o_top++ = o;
                continue;

            case OP_DOT:
                /*
                 * aggr key => value (os)
                 */
                if ((o = fetch(o_top[-2], o_top[-1])) == NULL)
                    goto fail;
                --o_top;
                o_top[-1] = o;
                goto stable_stacks_continue;

            case OP_DOTKEEP:
                /*
                 * aggr key => aggr key value (os)
                 */
                if ((o = fetch(o_top[-2], o_top[-1])) == NULL)
                    goto fail;
                *o_top++ = o;
                continue;

            case OP_DOTRKEEP:
                /*
                 * aggr key => value aggr key value (os)
                 *
                 * Used in postfix ++/-- for value.
                 */
                if ((o = fetch(o_top[-2], o_top[-1])) == NULL)
                    goto fail;
                o_top += 2;
                o_top[-1] = o;
                o_top[-2] = o_top[-3];
                o_top[-3] = o_top[-4];
                o_top[-4] = o;
                continue;

            case OP_ASSIGN:
                /*
                 * aggr key value => - (os, for effect)
                 *                => value (os, for value)
                 *                => aggr key (os, for lvalue)
                 */
                {
                    register object_t   *k;

                    if
                    (
                        isstring(k = o_top[-2])
                        &&
                        stringof(k)->s_vsver == vsver
                        &&
                        objof(stringof(k)->s_struct) == o_top[-3]
                    )
                    {
                        if (objof(stringof(k)->s_struct)->o_flags & O_ATOM)
                        {
                            error = "attempt to write on constant struct";
                            goto fail;
                        }
#ifndef NODEBUGGING
                        if (ici_debug_enabled)
                            ici_debug->idbg_watch(o_top[-3], k, o_top[-1]);
#endif
                        stringof(k)->s_slot->sl_value = o_top[-1];
                    }
                    else
                    {
#ifndef NODEBUGGING
                        if (ici_debug_enabled)
                            ici_debug->idbg_watch(o_top[-3], k, o_top[-1]);
#endif
                        if (assign(o_top[-3], k, o_top[-1]))
                            goto fail;
                    }
                }
                switch (opof(o)->op_code)
                {
                case FOR_EFFECT:
                    o_top -= 3;
                    break;

                case FOR_VALUE:
                    o_top[-3] = o_top[-1];
                    o_top -= 2;
                    break;

                case FOR_LVALUE:
                    --o_top;
                    break;
                }
                continue;

            case OP_SWAP:
                /*
                 * aggr1 key1 aggr2 key2        =>
                 *                              => value1
                 *                              => aggr1 key1
                 */
                {
                    register object_t   *v1;
                    register object_t   *v2;

                    if ((v1 = fetch(o_top[-4], o_top[-3])) == NULL)
                        goto fail;
                    if ((v2 = fetch(o_top[-2], o_top[-1])) == NULL)
                        goto fail;
                    incref(v2);
                    if (assign(o_top[-2], o_top[-1], v1))
                    {
                        decref(v2);
                        goto fail;
                    }
                    if (assign(o_top[-4], o_top[-3], v2))
                    {
                        decref(v2);
                        goto fail;
                    }
                    decref(v2);
                    switch (opof(o)->op_code)
                    {
                    case FOR_EFFECT:
                        o_top -= 4;
                        break;

                    case FOR_VALUE:
                        o_top[-4] = v2;
                        o_top -= 3;
                        break;

                    case FOR_LVALUE:
                        o_top -= 2;
                        break;
                    }
                }
                continue;

            case OP_IFNOTBREAK:
                /*
                 * bool => - (os)
                 *      => [o_break] (xs)
                 */
                if (!isfalse(o_top[-1]))
                {
                    --o_top;
                    continue;
                }
                --o_top;
                /* Falling through. */
            case OP_BREAK:
                /*
                 * Pop the execution stack until a looper or switcher
                 * is found and disgard it (and the thing under it).
                 */
                {
                    register object_t   **s;

                    for (s = x_top; s > xs->a_base + 1; --s)
                    {
                        if (s[-1] == objof(&frame))
                            break;
                        if (s[-1] == objof(&o_looper)
                         || s[-1] == objof(&o_switcher))
                        {
                            x_top = s - 2;
                            goto cont;
                        }
                        if (isforall(s[-1]))
                        {
                            x_top = s - 1;
                            goto cont;
                        }
                    }
                }
                error = "break not within loop or switch";
                goto fail;

            case OP_LOOP:
                /*
                 *      => obj looper pc (xs)
                 * obj  => - (os)
                 */
                *x_top++ = o_top[-1];
                *x_top++ = objof(&o_looper);
                /* Fall through */
            case OP_EXEC:
                /*
                 * array => - (os)
                 *       => pc (xs)
                 */
                /*
                 * In-line expansion of new_pc(o_top[-1]) (from pc.c).
                 */
                if ((o = objof(talloc(pc_t))) == NULL)
                    goto fail;
                o->o_type = &pc_type;
                o->o_tcode = TC_PC;
                o->o_flags = 0;
                o->o_nrefs = 0;
                rego(o);
                --o_top;
                pcof(o)->pc_code = arrayof(*o_top);
                pcof(o)->pc_next = arrayof(*o_top)->a_base;
                pcof(o)->pc_limit = arrayof(*o_top)->a_top;
                *x_top++ = o;
                continue;

            case OP_BINOP:
#ifndef BINOPFUNC
#include        "binop.h"
#else
                if (ici_op_binop(o))
                    goto fail;
#endif
                goto stable_stacks_continue;
            }
            continue;
        }

    fail:
        {
            catch_t     *c;

            if (error == NULL)
                error = "error";
            if ((c = ici_unwind()) == NULL || c == frame)
                goto badfail;
            incref(c);
            if (ici_cmkvar(structof(v_top[-1]), "error", 's', error))
            {
                decref(c);
                goto badfail;
            }
            if ((*x_top = objof(new_pc(arrayof(c->c_catcher)))) == NULL)
                goto badfail;
            decref(c);
            ++x_top;
            continue;

        badfail:
#ifndef NODEBUGGING
            if (ici_debug_enabled)
                ici_debug->idbg_error(error, src);
#endif
            expand_error(src->s_lineno, src->s_filename);
            decref(src);
            return NULL;
        }

    cont:;

    }
}

op_t    o_quote         = {OBJ(TC_OP, op_type), NULL, OP_QUOTE};
