#include "func.h"
#include "exec.h"
#include "ptr.h"
#include "struct.h"
#include "op.h"
#include "pc.h"
#include "str.h"
#include "catch.h"
#include "buf.h"
#include "mark.h"
#include "null.h"
#include "primes.h"
#ifndef NOPROFILE
#include "profile.h"
#endif


/*
 * Cached references to strings.
 */
static string_t *string_vargs;  /* Name "vargs". */


/*
 * Cleans up data structures allocated/referenced in this module.
 * Required for a clean shutdown.
 */
void
uninit_func(void)
{
    if (string_vargs != NULL)
    {
        decref(string_vargs);
        string_vargs = NULL;
    }
}


string_t *
func_name(func_t *f)
{
    if (objof(f)->o_flags & O_CFUNC)
        return get_cname(cfuncof(f)->cf_name);
    else
        return f->f_name;
}

/*
 * Mark this and referenced unmarked objects, return memory costs.
 * See comments on t_mark() in object.h.
 */
static unsigned long
mark_func(object_t *o)
{
    long        mem;

    o->o_flags |= O_MARK;
    if (o->o_flags & O_CFUNC)
        return sizeof(func_t);
    mem = sizeof(func_t);
    if (funcof(o)->f_code != NULL)
        mem += mark(objof(funcof(o)->f_code));
    if (funcof(o)->f_args != NULL)
        mem += mark(objof(funcof(o)->f_args));
    if (funcof(o)->f_autos != NULL)
        mem += mark(objof(funcof(o)->f_autos));
    if (funcof(o)->f_name != NULL)
        mem += mark(objof(funcof(o)->f_name));
    return mem;
}

/*
 * Returns 0 if these objects are equal, else non-zero.
 * See the comments on t_cmp() in object.h.
 */
static int
cmp_func(object_t *o1, object_t *o2)
{
    if ((o1->o_flags & O_CFUNC) || (o2->o_flags & O_CFUNC))
        return o1 != o2;
    return funcof(o1)->f_code != funcof(o2)->f_code
        || funcof(o1)->f_autos != funcof(o2)->f_autos
        || funcof(o1)->f_args != funcof(o2)->f_args
        || funcof(o1)->f_name != funcof(o2)->f_name;
}

/*
 * Return a hash sensitive to the value of the object.
 * See the comment on t_hash() in object.h
 */
static unsigned long
hash_func(object_t *o)
{
    if (o->o_flags & O_CFUNC)
        return (unsigned long)o * FUNC_PRIME;
    return (unsigned long)funcof(o)->f_code * FUNC_PRIME;
}

func_t *
new_func()
{
    register func_t     *f;

    if ((f = talloc(func_t)) == NULL)
        return NULL;
    memset((char *)f, 0, sizeof(func_t));
    objof(f)->o_type = &func_type;
    objof(f)->o_nrefs = 1;
    rego(f);
    return f;
}

/*
 * Return the object which at key k of the obejct o, or NULL on error.
 * See the comment on t_fetch in object.h.
 */
static object_t *
fetch_func(object_t *o, object_t *k)
{
    object_t            *r;

    error = NULL;
    r = NULL;
    if (o->o_flags & O_CFUNC)
    {
        if (k == objof(get_cname("name")))
            r = objof(get_cname(cfuncof(o)->cf_name));
    }
    else
    {
        if (k == objof(get_cname("autos")))
            r = objof(funcof(o)->f_autos);
        else if (k == objof(get_cname("args")))
            r = objof(funcof(o)->f_args);
        else if (k == objof(get_cname("name")))
            r = objof(funcof(o)->f_name);
    }
    if (r == NULL && error == NULL)
        r = objof(&o_null);
    return r;
}


/*
 * ici_assign_cfuncs
 *
 * Assign the structure s all the intrinsic functions listed in the array
 * of cfunc_t structures pointed to by cf. The array must be terminated by
 * an entry with a cf_name of NULL. Typically, entries in the array are
 * formated as:
 *
 *  {CF_OBJ,    "func",     f_func},
 *
 * Where CF_OBJ is a convenience macro from func.h to take care of the
 * normal object header, "func" is the name your function will be assigned
 * to in the given struct, and f_func is a C function obeying the rules
 * of ICI intrinsic functions.
 *
 * Returns non-zero on error, in which case error is set, else zero.
 */
int
ici_assign_cfuncs(struct_t *s, cfunc_t *cf)
{
    string_t    *n;

    while (cf->cf_name != NULL)
    {
        cf->o_head.o_type = &func_type;
        if ((n = new_cname(cf->cf_name)) == NULL)
            return 1;
        if (assign(s, n, cf))
        {
            decref(n);
            return 1;
        }
        decref(n);
        ++cf;
    }
    return 0;
}

/*
 * def_cfuncs
 *
 * Define the given intrinsic functions in the current static scope.
 * See ici_assign_cfuncs() above for details.
 *
 * Returns non-zero on error, in which case error is set, else zero.
 */
int
def_cfuncs(cfunc_t *cf)
{
    return ici_assign_cfuncs(structof(v_top[-1])->s_super, cf);
}

int
ici_op_return()
{
    register object_t   **x;

    x = xs->a_top;
    while
    (
        --x >= xs->a_base
        &&
        !ismark(*x)
        &&
        !(iscatch(*x) && isnull(catchof(*x)->c_catcher))
    )
        ;
    if (x < xs->a_base || !ismark(*x))
    {
        error = "return not in function";
        return 1;
    }
    x_top = x;
    --v_top;
    NEXT_VSVER;
#ifndef NOPROFILE
    if (ici_profile_active)
        ici_profile_return();
#endif
    return 0;
}

/*
 * arg(N-1) .. arg1 arg0 func   =>                      (os)
 *                              => auto-struct          (vs)
 *                      call(N) => mark pc              (xs)
 *
 * Calling a function pushes a structure for auto variables on the
 * variable stack. It then pushes a mark and a pc starting at the first
 * element of the code array on the execution stack. Any arguments are
 * assigned to the corresponding formal argument names in the auto var
 * structure. The count of actual arguments is encoded in the call operator.
 */
int
ici_op_call()
{
    register func_t     *f;
    register struct_t   *d;     /* The local variable structure. */
    register pc_t       *pc;
    register object_t   **ap;   /* Actual parameter. */
    register object_t   **fp;   /* Formal parameter. */
    slot_t              *sl;
    array_t             *va;
    int                 n;
    object_t            *o;     /* "self" if is a pointer call, else NULL. */

    f = funcof(o_top[-1]);
    if (isptr(objof(f)))
    {
        o = ptrof(f)->p_aggr;
        if ((f = funcof(fetch(o, ptrof(f)->p_key))) == NULL)
            return 1;
    }
    else
        o = NULL;

    if (!isfunc(objof(f)))
    {
        char    n[30];

        sprintf(buf, "attempt to call %s", objname(n, objof(f)));
        error = buf;
        return 1;
    }

#ifndef NOPROFILE
    if (ici_profile_active)
        ici_profile_call(f);
#endif

    if (objof(f)->o_flags & O_CFUNC)
    {
        /*
         * This is an intrinsic function.
         */
        int result = (*cfuncof(f)->cf_cfunc)(o);
#ifndef NOPROFILE
        if (ici_profile_active)
            ici_profile_return();
#endif
        return result;
    }

    if ((pc = new_pc(f->f_code)) == NULL)
        return 1;
    incref(pc);
    if ((d = structof(copy(objof(f->f_autos)))) == NULL)
        goto fail;
    n = NARGS(); /* Number of actual args. */
    ap = ARGS();
    if (f->f_args != NULL)
    {
        /*
         * There are explicit formal parameters.
         */
        fp = f->f_args->a_base;
        if (o != NULL && fp < f->f_args->a_top)
        {
            /*
             * This is a ptr call. Insert the implicit object as the
             * first argument.
             */
            if (assign(d, *fp++, o))
                goto fail;
            o = NULL; /* Consume, won't go into vargs now. */
        }
        /*
         * Assign the explicit actuals to the explicit formals.
         */
        while (fp < f->f_args->a_top && n > 0)
        {
            if (assign(d, *fp++, *ap--))
                goto fail;
            --n;
        }
    }
    if (o != NULL)
        ++n;
    va = NULL;
    if
    (
        n > 0
        &&
        (string_vargs != NULL || (string_vargs = new_cname("vargs")) != NULL)
        &&
        (sl = find_raw_slot(d, objof(string_vargs))) != NULL
        &&
        (va = new_array()) != NULL
        &&
        pushcheck(va, n) == 0
    )
    {
        /*
         * There are left-over actual parameters and a "vargs"
         * auto to put them in, and everything else looks good.
         */
        if (o != NULL)
        {
            *va->a_top++ = o;
            --n;
        }
        while (--n >= 0)
            *va->a_top++ = *ap--;
        sl->sl_value = objof(va);
    }
    if (va != NULL)
        decref(va);

    *v_top++ = objof(d);
    decref(d);
    NEXT_VSVER;
    o_top -= NARGS() + 1;
    x_top[-1] = objof(&o_mark);
    *x_top++ = objof(pc);
    decref(pc);
    return 0;

fail:
    decref(pc);
    if (d != NULL)
        decref(d);
    return 1;
}

type_t  func_type =
{
    mark_func,
    free_simple,
    hash_func,
    cmp_func,
    copy_simple,
    assign_simple,
    fetch_func,
    "func"
};

op_t    o_return        = {OBJ(TC_OP, op_type), ici_op_return};
