#include "exec.h"
#include "buf.h"
#include "int.h"
#include "str.h"
#include "float.h"
#include "func.h"
#include "primes.h"

#include <limits.h>

#define VERBOSE 0

/*
 * The global error message pointer.
 */
char            *error;

/*
 * All objects are in the objects list or completely static and
 * known to never require collection.
 */
object_t        **objs;         /* List of all objects. */
object_t        **objs_limit;   /* First element we can't use in list. */
object_t        **objs_top;     /* Next unused element in list. */

object_t        **atoms;        /* Hash table of atomic objects. */
int             atomsz;         /* Number of slots in hash table. */
static int      natoms;         /* Number of atomic objects. */

/*
 * Format a human readable version of the object in 30 chars or less.
 */
char *
objname(char *p, object_t *o)
{
#define OBJNAME_BUF_SIZE 30

    if (isstring(o))
    {
        if (stringof(o)->s_nchars > 10)
            sprintf(p, "\"%.10s...\"", stringof(o)->s_chars);
        else
            sprintf(p, "\"%s\"", stringof(o)->s_chars);
    }
    else if (isint(o))
        sprintf(p, "%ld", intof(o)->i_value);
    else if (isfloat(o))
        sprintf(p, "%g", floatof(o)->f_value);
    else if (isfunc(o))
    {
        string_t *name = func_name(funcof(o));

        /* Limit the output to the 30 characters available. */
        if (name->s_nchars > OBJNAME_BUF_SIZE - 2 - 1)
            sprintf(p, "%.*s...()", OBJNAME_BUF_SIZE - 6, name->s_chars);
        else
            sprintf(p, "%.*s()", name->s_nchars, name->s_chars);
    }
    else if (strchr("aeiou", o->o_type->t_name[0]) != NULL)
        sprintf(p, "an %s", o->o_type->t_name);
    else
        sprintf(p, "a %s", o->o_type->t_name);
    return p;
}

/*
 * Free this object and associated memory (but not other objects).
 * See the comments on t_free() in object.h.
 */
void
free_simple(object_t *o)
{
    ici_free(o);
}

/*
 * Return a copy of the given object, or NULL on error.
 * See the comment on t_copy() in object.h.
 */
object_t *
copy_simple(object_t *o)
{
    incref(o);
    return o;
}

/*
 * Assign to key k of the object o the value v. Return 1 on error, else 0.
 * See the comment on t_assign() in object.h.
 *
 * This is a generic function which can be used directly as the assign
 * function of a type which doesn't support assignment, or called from
 * within a custom assign function in cases where the particular assignment
 * is illegal.
 */
int
assign_simple(object_t *o, object_t *k, object_t *v)
{
    char        n1[30];
    char        n2[30];
    char        n3[30];

    sprintf(buf, "attempt to set %s keyed by %s to %s",
        objname(n1, o),
        objname(n2, k),
        objname(n3, v));
    error = buf;
    return 1;
}

/*
 * Return the object at key k of the obejct o, or NULL on error.
 * See the comment on t_fetch in object.h.
 *
 * This is a generic function which can be used directly as the fetch
 * function of a type which doesn't support fetching, or called from
 * within a custom fetch function in cases where the particular fetch
 * is illegal.
 */
object_t *
fetch_simple(object_t *o, object_t *k)
{
    char        n1[30];
    char        n2[30];

    sprintf(buf, "attempt to read %s keyed by %s",
        objname(n1, o),
        objname(n2, k));
    error = buf;
    return NULL;
}

/*
 * Returns 0 if these objects are equal, else non-zero.
 * For objects which can't be copied and are intrinsically unique.
 * See the comments on t_cmp() in object.h.
 */
int
cmp_unique(object_t *o1, object_t *o2)
{
    return o1 != o2;
}

/*
 * Return a hash sensitive to the value of the object.
 * See the comment on t_hash() in object.h
 */
unsigned long
hash_unique(object_t *o)
{
    return ((unsigned long)o >> 4) * UNIQUE_PRIME;
}

/*
 * Grow the hash table of atoms to be four times as big, plus three.
 */
static void
grow_atoms(void)
{
    register object_t   **o;
    register int        i;
    object_t            **olda;
    unsigned long       newz;

    newz = atomsz * 2;
    if ((o = (object_t **)ici_alloc(newz * sizeof(object_t *))) == NULL)
        return;
    i = atomsz;
    atomsz = newz;
    memset((char *)o, 0, newz * sizeof(object_t *));
    natoms = 0;
    olda = atoms;
    atoms = o;
    while (--i >= 0)
    {
        if (olda[i] != NULL)
        {
            olda[i]->o_flags &= ~O_ATOM;
            atom(olda[i], 1);
        }
    }
    ici_free((char *)olda);
}

/*
 * Return an object equal to the one given, but possibly shared by others.
 * Never fails, at worst it just returns its argument.  If the lone flag
 * is given, the object is free'd if it isn't used.  ("lone" because the
 * caller has the lone reference to it and will replace that with what
 * atom returns anyway.)  If the lone flag is not given, and the object
 * would be used, a copy will be used.  Also note that if lone is true and
 * the object is not used, the nrefs of the passed object will be transfered
 * to the object being returned.
 */
object_t *
atom(object_t *o, int lone)
{
    object_t    **po;

    if (o->o_flags & O_ATOM)
        return o;
    for
    (
        po = &atoms[ici_atom_hash_index(hash(o))];
        *po != NULL;
        --po < atoms ? po = atoms + atomsz - 1 : NULL
    )
    {
        if (o->o_type == (*po)->o_type && cmp(o, *po) == 0)
        {
            if (lone)
            {
                (*po)->o_nrefs += o->o_nrefs;
                o->o_nrefs = 0;
            }
            return *po;
        }
    }

    /*
     * Not found.  Add this object (or a copy of it) to the atom pool.
     */
    if (!lone)
    {
        if ((*po = copy(o)) == NULL)
            return o;
        o = *po;
    }
    *po = o;
    o->o_flags |= O_ATOM;
    if (++natoms > atomsz / 2)
        grow_atoms();
    if (!lone)
        decref(o);
    return o;
}

/*
 * Probe the atom pool for an atomic form of o.  If found, return that
 * atomic form, else NULL.  Used by various new_*() routines.  These
 * routines generally set up a dummy version of the object being made
 * which is passed to this probe.  If it finds a match, that is returned,
 * thus avoiding the allocation of an object that will may be thrown
 * away anyway.
 */
object_t *
atom_probe(object_t *o)
{
    object_t    **po;

    for
    (
        po = &atoms[ici_atom_hash_index(hash(o))];
        *po != NULL;
        --po < atoms ? po = atoms + atomsz - 1 : NULL
    )
    {
        if (o->o_type == (*po)->o_type && cmp(o, *po) == 0)
            return *po;
    }
    return NULL;
}

/*
 * Quick search for an int to save allocation/deallocation if it already
 * exists.
 */
int_t *
atom_int(long i)
{
    object_t    *o;
    object_t    **po;

    /*
     * NB: There is an in-line version of this code in binop.h
     */
    for
    (
        po = &atoms[ici_atom_hash_index((unsigned long)i * INT_PRIME)];
        (o = *po) != NULL;
        --po < atoms ? po = atoms + atomsz - 1 : NULL
    )
    {
        if (isint(o) && intof(o)->i_value == i)
            return intof(o);
    }
    return NULL;
}

static void
unatom(object_t *o)
{
    register object_t   **sl;
    register object_t   **ss;
    register object_t   **ws;   /* Wanted position. */

    for
    (
        ss = &atoms[ici_atom_hash_index(hash(o))];
        *ss != NULL;
        --ss < atoms ? ss = atoms + atomsz - 1 : NULL
    )
    {
        if (o == *ss)
           goto delete;
    }
    /*printf("Warning: could not find atom being deleted\n");*/
    return;

delete:
    o->o_flags &= ~O_ATOM;
    --natoms;
    sl = ss;
    /*
     * Scan "forward" bubbling up entries which would rather be at our
     * current empty slot.
     */
    for (;;)
    {
        if (--sl < atoms)
            sl = atoms + atomsz - 1;
        if (*sl == NULL)
            break;
        ws = &atoms[ici_atom_hash_index(hash(*sl))];
        if
        (
            (sl < ss && (ws >= ss || ws < sl))
            ||
            (sl > ss && (ws >= ss && ws < sl))
        )
        {
            /*
             * The value at sl, which really wants to be at ws, should go
             * into the current empty slot (ss).  Copy it to there and update
             * ss to be here (which now becomes empty).
             */
            *ss = *sl;
            ss = sl;
        }
    }
    *ss = NULL;
}

void
grow_objs(o)
object_t        *o;
{
    register object_t   **newobjs;
    register int        newz;

    newz = 2 * (objs_limit - objs) + 1;
    if ((newobjs = (object_t **)ici_alloc(newz * sizeof(object_t *))) == NULL)
        return;
    memcpy((char *)newobjs, (char *)objs, (char *)objs_limit - (char *)objs);
    objs_limit = newobjs + newz;
    objs_top = newobjs + (objs_top - objs);
    memset((char *)objs_top, 0, (char *)objs_limit - (char *)objs_top);
    ici_free((char *)objs);
    objs = newobjs;
    *objs_top++ = o;
}

/*
 * Mark sweep garbage collection.  Should be safe to do any time, as new
 * objects are created without the nrefs == 0 which allows them to be
 * collected.  They must be explicitly lost before they are subject
 * to garbage collection.  But of course all code must be careful not
 * to hang on to "found" objects where they are not accessible, or they
 * will be collected.  You can incref() them if you want.  All "held" objects
 * will cause all objects referenced from them to be marked (ie, not
 * collected), as long as they are registered on either the global object
 * list or in the atom pool.  Thus statically declared objects which
 * reference other objects (very rare) must be appropriately registered.
 */
void
collect(void)
{
    register object_t   **a;
    register object_t   *o;
    register object_t   **b;
    register int        ndead_atoms;
    register long       mem;    /* Total mem tied up in refed objects. */

    /*
     * Mark all objects which are referenced (and thus what they ref).
     */

    mem = 0;
    for (a = objs; a < objs_top; ++a)
    {
        if ((*a)->o_nrefs != 0)
            mem += mark(*a);
    }

    /*
     * Count how many atoms are going to be retained and how many are
     * going to be lost so we can decide on the fastest method.
     */
    ndead_atoms = 0;
    for (a = objs; a < objs_top; ++a)
    {
        if (((*a)->o_flags & (O_ATOM|O_MARK)) == O_ATOM)
            ++ndead_atoms;
    }

    /*
     * Collection phase.  Discard unmarked objects, compact down marked
     * objects and fix up the atom pool.
     *
     * Deleteing an atom from the atom pool is (say) once as expensive
     * as adding one.  Use this to determine which is quicker; rebuilding
     * the atom pool or deleting dead ones.
     */
    if (1 * ndead_atoms > (natoms - ndead_atoms))
    {
        /*
         * Rebuilding the atom pool is a better idea.
         */
        memset((char *)atoms, 0, atomsz * sizeof(object_t *));
        natoms = 0;
        b = objs;
        for (a = objs; a < objs_top; ++a)
        {
            if (((o = *a)->o_flags & O_MARK) == 0)
                freeo(o);
            else
            {
                o->o_flags &= ~O_MARK;
                *b++ = o;
                if (o->o_flags & O_ATOM)
                {
                    o->o_flags &= ~O_ATOM;
                    atom(o, 1);
                }
            }
        }
        objs_top = b;
    }
    else
    {
        /*
         * Faster to delete dead atoms as we go.
         */
        b = objs;
        for (a = objs; a < objs_top; ++a)
        {
            if (((o = *a)->o_flags & O_MARK) == 0)
            {
                if (o->o_flags & O_ATOM)
                    unatom(o);
                freeo(o);
            }
            else
            {
                o->o_flags &= ~O_MARK;
                *b++ = o;
            }
        }
        objs_top = b;
    }

/* printf("mem=%ld nobjects=%d natoms=%d\n", mem, objs_top - objs, natoms); */

    /*
     * The amount of memory tied up in referenced objects after a collect
     * helps control the pace of garbage collection.  We keep a nominal
     * minimum value to stop collection when we aren't using much memory.
     * We also limit our measure to stop collects being to infrequent.
     */
    if (mem < 1024 * 32)
        ici_old_mem = 1024 * 32;
    else if (mem > 1024 * 512)
        ici_old_mem = 1024 * 512;
    else
        ici_old_mem = mem;
    ici_new_mem = 0;
}

#ifdef  BUGHUNT
void
bughunt_incref(object_t *o)
{
    if ((unsigned char)objof(o)->o_nrefs == (unsigned char)0xFF)
    {
        printf("Oops: ref count overflow\n");
        abort();
    }
    if (++objof(o)->o_nrefs > 10)
    {
        printf("Warning: nrefs %d > 10\n", objof(o)->o_nrefs);
        fflush(stdout);
    }
}

void
bughunt_decref(object_t *o)
{
    if (--o->o_nrefs < 0)
    {
        printf("Oops: ref count underflow\n");
        abort();
    }
}
#endif

#ifdef  SMALL
/*
 * If SMALL is defined we use simple functions for these type vectored
 * operations. They are normally macros.
 */

unsigned long
mark(object_t *o)
{
    if (o->o_flags & O_MARK)
        return 0L;
    return (*o->o_type->t_mark)(o);
}

void
freeo(object_t *o)
{
    (*o->o_type->t_free)(o);
}

unsigned long
hash(object_t *o)
{
    return (*o->o_type->t_hash)(o);
}

int
cmp(object_t *o1, object_t *o2)
{
    return (*o1->o_type->t_cmp)(o1, o2);
}

object_t *
copy(object_t *o)
{
    return (*o->o_type->t_copy)(o);
}

object_t *
fetch(object_t *o, object_t *k)
{
    return (*o->o_type->t_fetch)(o, k);
}

int
assign(object_t *o, object_t *k, object_t *v)
{
    return (*o->o_type->t_assign)(o, k, v);
}

void
rego(object_t *o)
{
    if (objs_top < objs_limit)
        *objs_top++ = o;
    else
        grow_objs(o);
}
#endif
