#include "exec.h"
#include "ptr.h"
#include "struct.h"
#include "int.h"
#include "op.h"
#include "buf.h"
#include "primes.h"

/*
 * Mark this and referenced unmarked objects, return memory costs.
 * See comments on t_mark() in object.h.
 */
static unsigned long
mark_ptr(object_t *o)
{
    o->o_flags |= O_MARK;
    return sizeof(ptr_t) + mark(ptrof(o)->p_aggr) + mark(ptrof(o)->p_key);
}

/*
 * Returns 0 if these objects are equal, else non-zero.
 * See the comments on t_cmp() in object.h.
 */
static int
cmp_ptr(object_t *o1, object_t *o2)
{
    return ptrof(o1)->p_aggr != ptrof(o2)->p_aggr
    	|| ptrof(o1)->p_key != ptrof(o2)->p_key;
}

/*
 * Return a hash sensitive to the value of the object.
 * See the comment on t_hash() in object.h
 */
static unsigned long
hash_ptr(object_t *o)
{
    return (unsigned long)ptrof(o)->p_aggr * PTR_PRIME_0
	+ (unsigned long)ptrof(o)->p_key * PTR_PRIME_1;
}

/*
 * Return the object at key k of the obejct o, or NULL on error.
 * See the comment on t_fetch in object.h.
 *
 * Fetching a "sub element" of a pointer is the operation of pointer indexing.
 * It works for any pointer whoes key into its aggregate is an integer, and
 * of course the key must be an integer too.  The final key is the sum of the
 * two keys.  But if the key is zero, just do *ptr.
 */
static object_t *
fetch_ptr(object_t *o, object_t *k)
{
    if (k == objof(o_zero))
	return fetch(ptrof(o)->p_aggr, ptrof(o)->p_key);
    if (!isint(k) || !isint(ptrof(o)->p_key))
	return fetch_simple(o, k);
    if (ptrof(o)->p_key == objof(o_zero))
	incref(k);
    else if ((k = objof(new_int(intof(k)->i_value + intof(ptrof(o)->p_key)->i_value)))
	    == NULL)
	return NULL;
    o = fetch(ptrof(o)->p_aggr, k);
    decref(k);
    return o;
}

/*
 * Assign to key k of the object o the value v. Return 1 on error, else 0.
 * See the comment on t_assign() in object.h.
 *
 * See above comment.
 */
static int
assign_ptr(object_t *o, object_t *k, object_t *v)
{
    if (k == objof(o_zero))
	return assign(ptrof(o)->p_aggr, ptrof(o)->p_key, v);
    if (!isint(k) || !isint(ptrof(o)->p_key))
	return assign_simple(o, k, v);
    if (ptrof(o)->p_key == objof(o_zero))
	incref(k);
    else if ((k = objof(new_int(intof(k)->i_value + intof(ptrof(o)->p_key)->i_value)))
	    == NULL)
	return 1;
    if (assign(ptrof(o)->p_aggr, k, v))
    {
	decref(k);
	return 1;
    }
    decref(k);
    return 0;
}

ptr_t *
new_ptr(object_t *a, object_t *k)
{
    register ptr_t	*p;

    if ((p = talloc(ptr_t)) == NULL)
	return NULL;
    objof(p)->o_type = &ptr_type;
    objof(p)->o_tcode = TC_PTR;
    objof(p)->o_flags = 0;
    objof(p)->o_nrefs = 1;
    rego(p);
    p->p_aggr = a;
    p->p_key = k;
    return ptrof(atom(objof(p), 1));
}

/*
 * aggr key => ptr
 */
int
ici_op_mkptr()
{
    register object_t	*o;

    if ((o = objof(new_ptr(o_top[-2], o_top[-1]))) == NULL)
	return 1;
    o_top[-2] = o;
    decref(o);
    --o_top;
    --x_top;
    return 0;
}

/*
 * ptr => aggr key
 */
int
ici_op_openptr()
{
    register ptr_t	*p;
    char		n[30];

    if (!isptr(objof(p = ptrof(o_top[-1]))))
    {
	sprintf(buf, "pointer required, but %s given", objname(n, o_top[-1]));
	error = buf;
	return 1;
    }
    o_top[-1] = p->p_aggr;
    *o_top++ = p->p_key;
    --x_top;
    return 0;
}

/*
 * ptr => obj
 */
int
ici_op_fetch()
{
    register ptr_t	*p;
    register object_t	*o;
    char		n[30];

    if (!isptr(objof(p = ptrof(o_top[-1]))))
    {
	sprintf(buf, "pointer required, but %s given", objname(n, o_top[-1]));
	error = buf;
	return 1;
    }
    if ((o = fetch(p->p_aggr, p->p_key)) == NULL)
	return 1;
    o_top[-1] = o;
    --x_top;
    return 0;
}

type_t	ptr_type =
{
    mark_ptr,
    free_simple,
    hash_ptr,
    cmp_ptr,
    copy_simple,
    assign_ptr,
    fetch_ptr,
    "ptr"
};

op_t	o_mkptr		= {OBJ(TC_OP, op_type), ici_op_mkptr};
op_t	o_openptr	= {OBJ(TC_OP, op_type), ici_op_openptr};
op_t	o_fetch		= {OBJ(TC_OP, op_type), ici_op_fetch};
