#include "str.h"
#include "re.h"
#include "exec.h"
#include "op.h"
#include "buf.h"
#include "primes.h"

#define	ICI /* Cause PCRE's internal.h to add special stuff for ICI */
#include "pcre/internal.h"

/*
 * Called when the above regexp routines detect an error.
 * Space for the offsets to matched sub-expressions in regexps and a global
 * counter for the number of matches. PCRE says the number of offsets must be
 * a multiple of three. See pcre.3 for an explanation.
 */

int	re_bra[(NSUBEXP + 1) * 3];
int	re_nbra;

regexp_t *
new_regexp(string_t *s, int flags)
{
    regexp_t	*r;
    pcre	*re;
    pcre_extra	*rex = NULL;
    int		errofs;

    /* Special test for possible failure of new_cname() in lex.c */
    if (s == NULL)
	return NULL;
    re = pcre_compile(s->s_chars, flags, (const char **)&error, &errofs, NULL);
    if (re == NULL)
	return NULL;
    if (pcre_info(re, NULL, NULL) > NSUBEXP)
    {
	sprintf(buf, "too many subexpressions in regexp, limit is %d", NSUBEXP);
	error = buf;
	goto fail;
    }
    rex = pcre_study(re, 0, (const char **)&error);
    if (error != NULL)
	goto fail;
    /* Note rex can be NULL if no extra info required */
    if ((r = (regexp_t *)ici_alloc(sizeof(regexp_t))) == NULL)
	goto fail;
    objof(r)->o_type = &regexp_type;
    objof(r)->o_tcode = TC_REGEXP;
    objof(r)->o_flags = 0;
    objof(r)->o_nrefs = 1;
    r->r_re  = re;
    r->r_rex = rex;
    if ((r->r_pat = stringof(copy(s))) == NULL)
    {
	ici_free(r);	/* Okay, prior to rego() */
	goto fail;
    }
    decref(r->r_pat);
    rego(r);
    return regexpof(atom(objof(r), 1));

fail:
    if (rex != NULL)
	ici_free(rex);
    if (re != NULL)
	ici_free(re);
    return NULL;
}

/*
 * Mark this and referenced unmarked objects, return memory costs.
 * See comments on t_mark() in object.h.
 */
static unsigned long
mark_regexp(object_t *o)
{
    o->o_flags |= O_MARK;
    return sizeof(regexp_t) + mark(regexpof(o)->r_pat) + ((real_pcre *)regexpof(o)->r_re)->size;
}

/*
 * Free this object and associated memory (but not other objects).
 * See the comments on t_free() in object.h.
 */
static void
free_regexp(object_t *o)
{
    if (regexpof(o)->r_rex != NULL)
	ici_free(regexpof(o)->r_rex);
    ici_free(regexpof(o)->r_re);
    ici_free(o);
}

/*
 * Return a hash sensitive to the value of the object.
 * See the comment on t_hash() in object.h
 */
static unsigned long
hash_regexp(object_t *o)
{
    /* static unsigned long	primes[] = {0xBF8D, 0x9A4F, 0x1C81, 0x6DDB}; */
    return (unsigned long)regexpof(o)->r_pat * 0x9A4F;
}

/*
 * Returns 0 if these objects are equal, else non-zero.
 * See the comments on t_cmp() in object.h.
 */
static int
cmp_regexp(object_t *o1, object_t *o2)
{
    return cmp(regexpof(o1)->r_pat, regexpof(o2)->r_pat);
}

static string_t	*string_pattern = NULL;
static string_t	*string_options = NULL;

static object_t *
fetch_regexp(object_t *o, object_t *k)
{
    if (string_pattern == NULL && (string_pattern = new_cname("pattern")) == NULL)
	return NULL;
    if (string_options == NULL && (string_options = new_cname("options")) == NULL)
	return NULL;
    if (k == objof(string_pattern))
	return objof(regexpof(o)->r_pat);
    if (k == objof(string_options))
    {
	int	options;
	pcre_info(regexpof(o)->r_re, &options, NULL);
	return objof(new_int(options));
    }
    return fetch_simple(o, k);
}

type_t	regexp_type =
{
    mark_regexp,
    free_regexp,
    hash_regexp,
    cmp_regexp,
    copy_simple,
    assign_simple,
    fetch_regexp,
    "regexp"
};
