#include "struct.h"
#include "ptr.h"
#include "exec.h"
#include "func.h"
#include "op.h"
#include "int.h"
#include "buf.h"
#include "str.h"
#include "pc.h"
#include "primes.h"

long    vsver   = 1;    /* Version number of the variable stack. */

#define HASHINDEX(k, s) (((unsigned long)(k) >> 6) & ((s)->s_nslots - 1))

/*
 * Find the struct slot which does, or should, contain the key k.
 * Searches up the super chain of the struct for a non-atmoic struct which
 * contains the key.  *sp is set to the particular struct the value was
 * found in, which  will always be the first struct if the value was not found.
 * That is: if the returned slot's sl_key == NULL, *sp is unchanged.
 */
static slot_t *
find_write_slot(struct_t **sp, object_t *k)
{
    register struct_t   *s;
    register slot_t     *slbase;        /* Slot in 1st super chain struct. */
    register slot_t     *slother;       /* Slot in other super chain structs. */

    s = *sp;
    /*
     * First search the first struct in the super chain.  If it is found
     * here, return it.  But if it is not we fall through with slbase set
     * to the slot in the first struct which should hold this value
     * (in case it is not found anywhere in the super chain).
     */
    slbase = &s->s_slots[HASHINDEX(k, s)];
    while (slbase->sl_key != NULL)
    {
        if (slbase->sl_key == k)
            return slbase;
        if (--slbase < s->s_slots)
            slbase = s->s_slots + s->s_nslots - 1;
    }

    /*
     * Continue searching with structs in the super chain (if any).  If the
     * slot is found, return it and patch *sp to be the struct it was found
     * in.  Else we will fall out and return the base slot without changing
     * the struct from the initial one.
     */
    while ((s = s->s_super) != NULL)
    {
        if (objof(s)->o_flags & O_ATOM)
            continue;
        slother = &s->s_slots[HASHINDEX(k, s)];
        while (slother->sl_key != NULL)
        {
            if (slother->sl_key == k)
            {
                *sp = s;
                return slother;
            }
            if (--slother < s->s_slots)
                slother = s->s_slots + s->s_nslots - 1;
        }
    }

    return slbase;
}

/*
 * Find the struct slot which does, or should, contain the key k.
 * Searches up the super chain of the struct if necessary.  *sp is
 * set to the particular struct the value was found in, which  will
 * always be the first struct if the value was not found.
 * That is: if the returned slot's sl_key == NULL, *sp is unchanged.
 */
slot_t *
find_slot(struct_t **sp, object_t *k)
{
    register struct_t   *s;
    register slot_t     *slbase;        /* Slot in 1st super chain struct. */
    register slot_t     *slother;       /* Slot in other super chain structs. */

    s = *sp;
    /*
     * First search the first struct in the super chain.  If it is found
     * here, return it.  But if it is not we fall through with slbase set
     * to the slot in the first struct which should hold this value
     * (in case it is not found anywhere in the super chain.
     */
    slbase = &s->s_slots[HASHINDEX(k, s)];
    while (slbase->sl_key != NULL)
    {
        if (slbase->sl_key == k)
            return slbase;
        if (--slbase < s->s_slots)
            slbase = s->s_slots + s->s_nslots - 1;
    }

    /*
     * Continue searching with structs in the super chain (if any).  If the
     * slot is found, return it and patch *sp to be the struct it was found
     * in.  Else we will fall out and return the base slot without changing
     * the struct from the initial one.
     */
    while ((s = s->s_super) != NULL)
    {
        slother = &s->s_slots[HASHINDEX(k, s)];
        while (slother->sl_key != NULL)
        {
            if (slother->sl_key == k)
            {
                *sp = s;
                return slother;
            }
            if (--slother < s->s_slots)
                slother = s->s_slots + s->s_nslots - 1;
        }
    }

    return slbase;
}

/*
 * Find the struct slot which does, or should, contain the key k.  Does
 * not look down the super chain.
 */
slot_t *
find_raw_slot(struct_t *s, object_t *k)
{
    register slot_t     *sl;

    sl = &s->s_slots[HASHINDEX(k, s)];
    while (sl->sl_key != NULL)
    {
        if (sl->sl_key == k)
            return sl;
        if (--sl < s->s_slots)
            sl = s->s_slots + s->s_nslots - 1;
    }
    return sl;
}

/*
 * Mark this and referenced unmarked objects, return memory costs.
 * See comments on t_mark() in object.h.
 */
static unsigned long
mark_struct(object_t *o)
{
    register slot_t     *sl;
    long                mem;

    o->o_flags |= O_MARK;
    mem = sizeof(struct_t) + structof(o)->s_nslots * sizeof(slot_t);
    if (structof(o)->s_super != NULL)
        mem += mark(structof(o)->s_super);
    if (structof(o)->s_nels == 0)
        return mem;
    for
    (
        sl = &structof(o)->s_slots[structof(o)->s_nslots - 1];
        sl >= structof(o)->s_slots;
        --sl
    )
    {
        if (sl->sl_key != NULL)
            mem += mark(sl->sl_key);
        if (sl->sl_value != NULL)
            mem += mark(sl->sl_value);
    }
    return mem;
}

/*
 * Free this object and associated memory (but not other objects).
 * See the comments on t_free() in object.h.
 */
static void
free_struct(object_t *o)
{
    if (structof(o)->s_slots != NULL)
        ici_free(structof(o)->s_slots);
    ici_free(o);
}

struct_t *
new_struct(void)
{
    register struct_t   *s;

    /*
     * NB: there is a copy of this sequence in copy_struct.
     */
    if ((s = (struct_t *)talloc(struct_t)) == NULL)
        return NULL;
    objof(s)->o_type = &struct_type;
    objof(s)->o_tcode = TC_STRUCT;
    objof(s)->o_flags = 0;
    objof(s)->o_nrefs = 1;
    rego(s);
    s->s_super = NULL;
    s->s_nels = 0;
    s->s_nslots = 4;
    /*
     * Note the 3 below is choosen to correspond with the smaller of
     * the special memory lists.
     */
    if ((s->s_slots = (slot_t *)ici_alloc(4 * sizeof(slot_t))) == NULL)
    {
        decref(s);      /* for garbage collection */
        return NULL;
    }
    memset(s->s_slots, 0, 4 * sizeof(slot_t));
    return s;
}

/*
 * Returns 0 if these objects are equal, else non-zero.
 * See the comments on t_cmp() in object.h.
 */
static int
cmp_struct(object_t *o1, object_t *o2)
{
    register int        i;
    register slot_t     *sl1;
    register slot_t     *sl2;

    if (structof(o1) == structof(o2))
        return 0;
    if (structof(o1)->s_nels != structof(o2)->s_nels)
        return 1;
    if (structof(o1)->s_super != structof(o2)->s_super)
        return 1;
    sl1 = structof(o1)->s_slots;
    i = structof(o1)->s_nslots;
    while (--i >= 0)
    {
        if (sl1->sl_key != NULL)
        {
            sl2 = find_raw_slot(structof(o2), sl1->sl_key);
            if (sl1->sl_key != sl2->sl_key || sl1->sl_value != sl2->sl_value)
                return 1;
        }
        ++sl1;
    }
    return 0;
}

/*
 * Return a hash sensitive to the value of the object.
 * See the comment on t_hash() in object.h
 */
static unsigned long
hash_struct(object_t *o)
{
    int                         i;
    unsigned long               hk;
    unsigned long               hv;
    slot_t                      *sl;

    hk = 0;
    hv = 0;
    sl = structof(o)->s_slots;
    i = structof(o)->s_nels;
    /*
     * This assumes NULL will become zero when cast to unsigned long.
     */
    while (--i >= 0)
    {
        hk += (unsigned long)sl->sl_key >> 4;
        hv += (unsigned long)sl->sl_value >> 4;
        ++sl;
    }
    return hv * STRUCT_PRIME_0 + hk * STRUCT_PRIME_1 + STRUCT_PRIME_2;
}

/*
 * Return a copy of the given object, or NULL on error.
 * See the comment on t_copy() in object.h.
 */
static object_t *
copy_struct(object_t *o)
{
    struct_t    *s;
    struct_t    *ns;

    s = structof(o);
    if ((ns = (struct_t *)talloc(struct_t)) == NULL)
        return NULL;
    objof(ns)->o_type = &struct_type;
    objof(ns)->o_tcode = TC_STRUCT;
    objof(ns)->o_flags = 0;
    objof(ns)->o_nrefs = 1;
    rego(ns);
    ns->s_super = s->s_super;
    ns->s_nels = 0;
    ns->s_nslots = 0;
    if ((ns->s_slots = (slot_t *)ici_alloc(s->s_nslots * sizeof(slot_t))) == NULL)
        goto fail;
    memcpy((char *)ns->s_slots, (char *)s->s_slots, s->s_nslots*sizeof(slot_t));
    ns->s_nels = s->s_nels;
    ns->s_nslots = s->s_nslots;
    return objof(ns);

fail:
    decref(ns);
    return NULL;
}

/*
 * Grow the struct s so that it has twice as many slots.
 */
static int
grow_struct(struct_t *s)
{
    register slot_t     *sl;
    register slot_t     *oldslots;
    register int        i;

    NEXT_VSVER;
    i = (s->s_nslots * 2 /*+ 1*/) * sizeof(slot_t);
    if ((sl = (slot_t *)ici_alloc(i)) == NULL)
        return 1;
    memset((char *)sl, 0, i);
    oldslots = s->s_slots;
    s->s_slots = sl;
    i = s->s_nslots;
    s->s_nslots *= 2;
    /*s->s_nslots += 1;*/
    while (--i >= 0)
    {
        if (oldslots[i].sl_key != NULL)
            *find_raw_slot(s, oldslots[i].sl_key) = oldslots[i];
    }
    ici_free((char *)oldslots);
    return 0;
}

/*
 * Set the value of key k in the struct s to the value v.  Will add the
 * entry if necessary and grow the struct if necessary.  Returns 1 on
 * failure, else 0.
 * See the comment on t_assign() in object.h.
 */
static int
assign_struct(object_t *o, object_t *k, object_t *v)
{
    register slot_t     *sl;

    if (o->o_flags & O_ATOM)
    {
        error = "attempt to modify an atomic struct";
        return 1;
    }
    if
    (
        isstring(k)
        &&
        stringof(k)->s_vsver == vsver
        &&
        stringof(k)->s_struct == structof(o)
    )
    {
        stringof(k)->s_slot->sl_value = v;
        return 0;
    }
    sl = find_write_slot((struct_t **)&o, k);
    if (sl->sl_key == NULL)
    {
        /*
         * We must still be in the base struct, we never find new
         * entries in super structs.
         */
        NEXT_VSVER;
        if (++structof(o)->s_nels >= structof(o)->s_nslots - structof(o)->s_nslots / 5)
        {
            /*
             * This struct has become 80% full.  Grow it.
             */
            if (grow_struct(structof(o)))
                return 1;
            sl = find_write_slot((struct_t **)&o, k);
        }
        sl->sl_key = k;
    }
    sl->sl_value = v;
    return 0;
}

/*
 * Remove the key from the structure.
 */
void
unassign_struct(struct_t *s, object_t *k)
{
    register slot_t     *sl;
    register slot_t     *ss;
    register slot_t     *ws;    /* Wanted position. */

    /*
     * ###Should this allow removal of elements from super structs?
     * Assuming no for the moment.
     */
    if ((ss = find_raw_slot(s, k))->sl_key == NULL)
        return;
    --s->s_nels;
    sl = ss;
    /*
     * Scan "forward" bubbling up entries which would rather be at our
     * current empty slot.
     */
    for (;;)
    {
        if (--sl < s->s_slots)
            sl = s->s_slots + s->s_nslots - 1;
        if (sl->sl_key == NULL)
            break;
        ws = &s->s_slots[HASHINDEX(sl->sl_key, s)];
        if
        (
            (sl < ss && (ws >= ss || ws < sl))
            ||
            (sl > ss && (ws >= ss && ws < sl))
        )
        {
            /*
             * The value at sl, which really wants to be at ws, should go
             * into the current empty slot (ss).  Copy it to there and update
             * ss to be here (which now becomes empty).
             */
            *ss = *sl;
            ss = sl;
        }
    }
    ss->sl_key = NULL;
    ss->sl_value = NULL;
}

/*
 * Return the object at key k of the obejct o, or NULL on error.
 * See the comment on t_fetch in object.h.
 */
static object_t *
fetch_struct(object_t *o, object_t *k)
{
    slot_t      *sl;

    if
    (
        isstring(k)
        &&
        stringof(k)->s_vsver == vsver
        &&
        stringof(k)->s_slot >= structof(o)->s_slots
        &&
        stringof(k)->s_slot < structof(o)->s_slots + structof(o)->s_nslots
    )
        return stringof(k)->s_slot->sl_value;
    sl = find_slot((struct_t **)&o, k);
    return sl->sl_key == NULL ? objof(&o_null) : sl->sl_value;
}

type_t  struct_type =
{
    mark_struct,
    free_struct,
    hash_struct,
    cmp_struct,
    copy_struct,
    assign_struct,
    fetch_struct,
    "struct"
};

op_t    o_namelvalue    = {OBJ(TC_OP, op_type), NULL, OP_NAMELVALUE};
op_t    o_dot           = {OBJ(TC_OP, op_type), NULL, OP_DOT};
op_t    o_dotkeep       = {OBJ(TC_OP, op_type), NULL, OP_DOTKEEP};
op_t    o_dotrkeep      = {OBJ(TC_OP, op_type), NULL, OP_DOTRKEEP};
